import pathlib
from typing import Optional

from copernicusmarine.core_functions.models import (
    ResponseSubset,
    SplitOnTimeOption,
)
from copernicusmarine.core_functions.request_structure import (
    create_subset_request,
)
from copernicusmarine.core_functions.subset_split_on import (
    subset_split_on_function,
)
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@log_exception_and_exit
def subset_split_on(
    on_variables: bool = False,
    on_time: Optional[SplitOnTimeOption] = None,
    concurrent_processes: Optional[int] = None,
    **kwargs,
) -> list[ResponseSubset]:
    """
    Extract a subset of data from a specified dataset using given parameters
    and split the output files based on variable names, time intervals or both.
    By default, sequentially download the data.

    The datasetID is required and can be found via the ``describe`` command.
    Accept all the arguments from the subset.

    Parameters
    ----------
    on_variables : bool, optional
        If True, split the output files based on variable names, by default False.
    on_time : Optional[SplitOnTimeOption], optional
        If provided, split the output files based on specified time intervals, by default None.
    concurrent_processes : Optional[int], optional
        Number of concurrent processes to use for downloading data, by default None. Should be greater or equal to 1.
    **kwargs
        Additional keyword arguments for subset request creation. See :class:`copernicusmarine.subset` for detailed accepted parameters.


    Returns
    -------
    list[ResponseSubset]
        A description of the downloaded data and its destination for each downloaded files.

    """  # noqa

    if kwargs.get("variables") is not None:
        _check_type(kwargs.get("variables"), list, "variables")
    if kwargs.get("platform_ids") is not None:
        _check_type(kwargs.get("platform_ids"), list, "platform_ids")

    if kwargs.get("output_directory") is not None:
        kwargs["output_directory"] = pathlib.Path(
            kwargs.get("output_directory")  # type: ignore
        )
    if kwargs.get("credentials_file") is not None:
        kwargs["credentials_file"] = pathlib.Path(
            kwargs.get("credentials_file")  # type: ignore
        )

    subset_request = create_subset_request(
        **kwargs,
    )

    return subset_split_on_function(
        subset_request=subset_request,
        on_variables=on_variables,
        on_time=on_time,
        concurrent_processes=concurrent_processes,
    )


def _check_type(value, expected_type: type, name: str):
    if not isinstance(value, expected_type):
        raise TypeError(f"{name} must be of type {expected_type.__name__}")



import logging
import pathlib
from typing import Optional

from copernicusmarine.core_functions.credentials_utils import (
    RECOVER_YOUR_CREDENTIALS_MESSAGE,
    copernicusmarine_validate_and_get_user,
    credentials_file_builder,
)

logger = logging.getLogger("copernicusmarine")


def login_function(
    username: Optional[str],
    password: Optional[str],
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
    check_credentials_valid: bool,
    configuration_file: Optional[pathlib.Path],
) -> bool:
    if check_credentials_valid:
        logger.info("Checking if credentials are valid.")
        if copernicusmarine_validate_and_get_user(
            configuration_file, username, password
        ):
            return True
        else:
            return False
    credentials_file, has_been_aborted = credentials_file_builder(
        username=username,
        password=password,
        configuration_file_directory=configuration_file_directory,
        force_overwrite=force_overwrite,
    )
    if has_been_aborted:
        logger.info("No configuration file have been modified.")
        return False
    if credentials_file is not None:
        logger.info(f"Credentials file stored in {credentials_file}.")
        return True
    else:
        logger.error("Invalid credentials.")
        logger.info("No configuration file have been modified.")
        logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
    return False



import fnmatch
import importlib.util
import json
import logging
import pathlib
import re
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, Type, TypeVar, Union

import pandas as pd
from dateutil.tz import UTC
from pydantic import BaseModel, ValidationError, field_validator

from copernicusmarine.core_functions.credentials_utils import (
    get_and_check_username_password,
)
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    log_deprecated_message,
)
from copernicusmarine.core_functions.exceptions import (
    LonLatSubsetNotAvailableInOriginalGridDatasets,
    MutuallyExclusiveArguments,
    XYNotAvailableInNonOriginalGridDatasets,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_FILE_FORMAT,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    FileFormat,
    VerticalAxis,
)
from copernicusmarine.core_functions.utils import datetime_parser
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
    XParameters,
    YParameters,
)
from copernicusmarine.versioner import __version__ as copernicusmarine_version

logger = logging.getLogger("copernicusmarine")


MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS: dict[str, str] = {
    "maximum_latitude": "maximum_y",
    "minimum_latitude": "minimum_y",
    "maximum_longitude": "maximum_x",
    "minimum_longitude": "minimum_x",
}

SubsetRequest_ = TypeVar("SubsetRequest_", bound="SubsetRequest")


class SubsetRequest(BaseModel):
    dataset_id: str
    username: str
    dataset_version: Optional[str] = None
    dataset_part: Optional[str] = None
    variables: Optional[list[str]] = None
    minimum_x: Optional[float] = None
    maximum_x: Optional[float] = None
    minimum_y: Optional[float] = None
    maximum_y: Optional[float] = None
    minimum_depth: Optional[float] = None
    maximum_depth: Optional[float] = None
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS
    start_datetime: Optional[datetime] = None
    end_datetime: Optional[datetime] = None
    platform_ids: Optional[list[str]] = None
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    )
    output_filename: Optional[str] = None
    file_format: FileFormat = DEFAULT_FILE_FORMAT
    service: Optional[str] = None
    output_directory: pathlib.Path = pathlib.Path(".")
    overwrite: bool = False
    skip_existing: bool = False
    netcdf_compression_level: int = 0
    netcdf3_compatible: bool = False
    dry_run: bool = False
    raise_if_updating: bool = False
    disable_progress_bar: bool = False
    staging: bool = False
    chunk_size_limit: int = -1

    def update(self, new_dict: dict) -> "SubsetRequest":
        filtered_dict = {
            key: value
            for key, value in new_dict.items()
            if value is not None
            and not (isinstance(value, (list, tuple, str)) and not value)
        }
        data = self.model_dump(
            exclude_defaults=True,
            exclude_unset=True,
            exclude_none=True,
        )
        data.update(filtered_dict)
        return self.model_validate(data)

    @field_validator("start_datetime", "end_datetime", mode="before")
    @classmethod
    def parse_datetime(
        cls, v: Optional[Union[datetime, pd.Timestamp, str]]
    ) -> Optional[datetime]:
        if v is None or v == "":
            return None
        if isinstance(v, str):
            return datetime_parser(v)
        if isinstance(v, pd.Timestamp):
            return v.to_pydatetime().astimezone(UTC)
        return v.astimezone(UTC)

    @classmethod
    def from_file(
        cls: Type[SubsetRequest_],
        filepath: pathlib.Path,
        username: Optional[str] = None,
    ) -> SubsetRequest_:
        with open(filepath) as json_file:
            json_content = json.load(json_file)
        if username:
            json_content["username"] = username
        json_content.pop("credentials_file", None)
        json_content.pop("password", None)
        transformed_data = cls._transform_deprecated_options(json_content)
        try:
            return cls(**transformed_data)
        except ValidationError as e:
            raise ValueError(f"Invalid request in file {filepath}: {e}")

    @classmethod
    def _transform_deprecated_options(
        cls: Type[SubsetRequest_], data: dict[str, Any]
    ) -> dict[str, Any]:
        transformed = {}
        for key, val in data.items():
            if key in DEPRECATED_OPTIONS:
                deprecated_option = DEPRECATED_OPTIONS[key]
                if deprecated_option.old_name == deprecated_option.new_name:
                    log_deprecated_message(deprecated_option.old_name, None)
                else:
                    log_deprecated_message(
                        deprecated_option.old_name,
                        deprecated_option.new_name,
                    )
                if deprecated_option.do_not_pass:
                    continue
                new_key = deprecated_option.new_name
                transformed[new_key] = val
            elif key in MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS:
                new_key = MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS[key]
                transformed[new_key] = val
            else:
                transformed[key] = val

        return transformed

    def get_temporal_parameters(
        self, axis_coordinate_id_mapping: dict[str, str]
    ) -> TemporalParameters:
        return TemporalParameters(
            start_datetime=self.start_datetime,
            end_datetime=self.end_datetime,
            coordinate_id=axis_coordinate_id_mapping.get("t", "time"),
        )

    def get_geographical_parameters(
        self,
        axis_coordinate_id_mapping: dict[str, str],
        is_original_grid: bool = False,
    ) -> GeographicalParameters:
        return GeographicalParameters(
            x_axis_parameters=XParameters(
                minimum_x=self.minimum_x,
                maximum_x=self.maximum_x,
                coordinate_id=axis_coordinate_id_mapping.get("x", "longitude"),
            ),
            y_axis_parameters=YParameters(
                minimum_y=self.minimum_y,
                maximum_y=self.maximum_y,
                coordinate_id=axis_coordinate_id_mapping.get("y", "latitude"),
            ),
            projection="originalGrid" if is_original_grid else "lonlat",
        )

    def get_depth_parameters(
        self, axis_coordinate_id_mapping: dict[str, str]
    ) -> DepthParameters:
        return DepthParameters(
            minimum_depth=self.minimum_depth,
            maximum_depth=self.maximum_depth,
            vertical_axis=self.vertical_axis,
            coordinate_id=axis_coordinate_id_mapping.get("z", "depth"),
        )


def convert_motu_api_request_to_structure(
    motu_api_request: str,
    username: str,
) -> SubsetRequest:
    prefix = "python -m motuclient "
    string = motu_api_request.replace(prefix, "").replace("'", "")
    arguments = [
        substr.strip() for substr in string.split("--")[1:]
    ]  # for subsubstr in substr.split(" ", maxsplit=1)]
    arg_value_tuples = [
        tuple(substr.split(" ", maxsplit=1)) for substr in arguments
    ]
    motu_api_request_dict: dict[str, Any] = {}
    for arg, value in arg_value_tuples:
        if arg == "variable":
            # special case for variable, since it can have multiple values
            motu_api_request_dict.setdefault(arg, []).append(value)
        else:
            motu_api_request_dict[arg] = value
    subset_request = SubsetRequest(
        dataset_id="",
        username=username,
    )
    conversion_dict = {
        "product-id": "dataset_id",
        "latitude-min": "minimum_y",
        "latitude-max": "maximum_y",
        "longitude-min": "minimum_x",
        "longitude-max": "maximum_x",
        "depth-min": "minimum_depth",
        "depth-max": "maximum_depth",
        "date-min": "start_datetime",
        "date-max": "end_datetime",
        "variable": "variables",
    }
    subset_request = subset_request.update(
        {
            conversion_dict[key]: value
            for key, value in motu_api_request_dict.items()
            if key in conversion_dict.keys()
        }
    )
    return subset_request


def create_subset_request(
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[list[str]] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS,
    start_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    end_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    platform_ids: Optional[list[str]] = None,
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    ),
    output_filename: Optional[str] = None,
    file_format: Optional[FileFormat] = None,
    service: Optional[str] = None,
    request_file: Optional[pathlib.Path] = None,
    output_directory: Optional[pathlib.Path] = None,
    credentials_file: Optional[pathlib.Path] = None,
    motu_api_request: Optional[str] = None,
    overwrite: bool = False,
    skip_existing: bool = False,
    dry_run: bool = False,
    disable_progress_bar: bool = False,
    staging: bool = False,
    netcdf_compression_level: int = 0,
    netcdf3_compatible: bool = False,
    chunk_size_limit: int = 0,
    raise_if_updating: bool = False,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    alias_min_x: Optional[float] = None,
    alias_max_x: Optional[float] = None,
    alias_min_y: Optional[float] = None,
    alias_max_y: Optional[float] = None,
    minimum_x: Optional[float] = None,
    maximum_x: Optional[float] = None,
    minimum_y: Optional[float] = None,
    maximum_y: Optional[float] = None,
) -> SubsetRequest:
    if staging:
        logger.warning(
            "Detecting staging flag for subset command. "
            "Data will come from the staging environment."
        )

    if overwrite:
        if skip_existing:
            raise MutuallyExclusiveArguments("overwrite", "skip_existing")
    if request_file and not username and not credentials_file:
        with open(request_file) as json_file:
            json_content = json.load(json_file)
        if "username" in json_content:
            username = json_content["username"]
        if "password" in json_content:
            password = json_content["password"]
        if "credentials_file" in json_content:
            credentials_file = pathlib.Path(json_content["credentials_file"])

    username, password = get_and_check_username_password(
        username,
        password,
        credentials_file,
    )
    subset_request = SubsetRequest(
        dataset_id=dataset_id or "",
        username=username,
    )
    if request_file:
        subset_request = SubsetRequest.from_file(
            request_file, username=username
        )
    if motu_api_request:
        motu_api_subset_request = convert_motu_api_request_to_structure(
            motu_api_request, username=username
        )
        subset_request = subset_request.update(
            motu_api_subset_request.__dict__
        )
    if not subset_request.dataset_id:
        raise ValueError("Please provide a dataset id for a subset request.")
    if netcdf3_compatible:
        documentation_url = (
            f"https://toolbox-docs.marine.copernicus.eu"
            f"/en/v{copernicusmarine_version}/installation.html#dependencies"
        )
        assert importlib.util.find_spec("netCDF4"), (
            "To enable the NETCDF3_COMPATIBLE option, the 'netCDF4' "
            f"package is required. "
            f"Please see {documentation_url}."
        )
    (
        minimum_x_axis,
        maximum_x_axis,
        minimum_y_axis,
        maximum_y_axis,
    ) = get_geographical_inputs(
        minimum_longitude,
        maximum_longitude,
        minimum_latitude,
        maximum_latitude,
        minimum_x,
        maximum_x,
        minimum_y,
        maximum_y,
        dataset_part,
    )
    if dataset_part == "originalGrid" and (
        alias_max_x is not None
        or alias_min_x is not None
        or alias_max_y is not None
        or alias_min_y is not None
    ):
        logger.debug(
            "Because you are using an originalGrid dataset, we are considering"
            " the options -x, -X, -y, -Y to be in m/km, not in degrees."
        )

    request_update_dict = {
        "dataset_version": dataset_version,
        "dataset_part": dataset_part,
        "variables": variables,
        "minimum_x": (
            minimum_x_axis if minimum_x_axis is not None else alias_min_x
        ),
        "maximum_x": (
            maximum_x_axis if maximum_x_axis is not None else alias_max_x
        ),
        "minimum_y": (
            minimum_y_axis if minimum_y_axis is not None else alias_min_y
        ),
        "maximum_y": (
            maximum_y_axis if maximum_y_axis is not None else alias_max_y
        ),
        "minimum_depth": minimum_depth,
        "maximum_depth": maximum_depth,
        "vertical_axis": vertical_axis,
        "start_datetime": start_datetime,
        "end_datetime": end_datetime,
        "platform_ids": platform_ids,
        "output_filename": output_filename,
        "file_format": file_format,
        "service": service,
        "output_directory": output_directory,
        "chunk_size_limit": chunk_size_limit,
    }
    # To be able to distinguish between set and unset values
    if skip_existing:
        request_update_dict["skip_existing"] = skip_existing
    if overwrite:
        request_update_dict["overwrite"] = overwrite
    if netcdf_compression_level:
        request_update_dict[
            "netcdf_compression_level"
        ] = netcdf_compression_level
    if netcdf3_compatible:
        request_update_dict["netcdf3_compatible"] = netcdf3_compatible
    if coordinates_selection_method != DEFAULT_COORDINATES_SELECTION_METHOD:
        request_update_dict[
            "coordinates_selection_method"
        ] = coordinates_selection_method
    if raise_if_updating:
        request_update_dict["raise_if_updating"] = raise_if_updating
    if dry_run:
        request_update_dict["dry_run"] = dry_run
    if staging:
        request_update_dict["staging"] = staging
    if disable_progress_bar or dry_run:
        request_update_dict["disable_progress_bar"] = True

    return subset_request.update(request_update_dict)


def get_geographical_inputs(
    minimum_longitude: Optional[float],
    maximum_longitude: Optional[float],
    minimum_latitude: Optional[float],
    maximum_latitude: Optional[float],
    minimum_x: Optional[float],
    maximum_x: Optional[float],
    minimum_y: Optional[float],
    maximum_y: Optional[float],
    dataset_part: Optional[str],
) -> tuple[Optional[float], Optional[float], Optional[float], Optional[float]]:
    """
    Returns the geographical selection of the user.

    Parameters
    ----------
    minimum_longitude : float
        Minimum longitude of the area of interest. For lat/lon datasets.
    maximum_longitude : float
        Maximum longitude of the area of interest. For lat/lon datasets.
    minimum_latitude : float
        Minimum latitude of the area of interest. For lat/lon datasets.
    maximum_latitude : float
        Maximum latitude of the area of interest. For lat/lon datasets.
    minimum_x : float
        Minimum x coordinate of the area of interest. For "originalGrid" datasets.
    maximum_x : float
        Maximum x coordinate of the area of interest. For "originalGrid" datasets.
    minimum_y : float
        Minimum y coordinate of the area of interest. For "originalGrid" datasets.
    maximum_y : float
        Maximum y coordinate of the area of interest. For "originalGrid" datasets.
    dataset_part : str
        The part of the dataset to be used. If "originalGrid", the x and y coordinates
        should be the inputs.

    Returns
    -------
    tuple[Optional[float], Optional[float], Optional[float], Optional[float]]
        The geographical selection of the user. (minimum_x_axis, maximum_x_axis, minimum_y_axis, maximum_y_axis).

    Raises
    ------

    LonLatSubsetNotAvailableInOriginalGridDatasets
        If the dataset is "originalGrid" and the user tries to use lat/lon coordinates.
    XYNotAvailableInNonOriginalGridDatasets
        If the dataset is not "originalGrid" and the user tries to use x/y coordinates.
    """  # noqa: E501

    if dataset_part == "originalGrid":
        if (
            minimum_longitude is not None
            or maximum_longitude is not None
            or minimum_latitude is not None
            or maximum_latitude is not None
        ):
            raise LonLatSubsetNotAvailableInOriginalGridDatasets
        else:
            return (
                minimum_x,
                maximum_x,
                minimum_y,
                maximum_y,
            )
    else:
        if (
            minimum_x is not None
            or maximum_x is not None
            or minimum_y is not None
            or maximum_y is not None
        ):
            raise XYNotAvailableInNonOriginalGridDatasets
        else:
            return (
                minimum_longitude,
                maximum_longitude,
                minimum_latitude,
                maximum_latitude,
            )


@dataclass
class GetRequest:
    dataset_id: str
    dataset_url: Optional[str] = None
    dataset_version: Optional[str] = None
    dataset_part: Optional[str] = None
    no_directories: bool = False
    output_directory: str = "."
    overwrite: bool = False
    filter: Optional[str] = None
    regex: Optional[str] = None
    file_list: Optional[pathlib.Path] = None
    sync: bool = False
    sync_delete: bool = False
    index_parts: bool = False
    direct_download: Optional[list[str]] = None
    dry_run: bool = False
    skip_existing: bool = False

    def update(self, new_dict: dict):
        """Method to update values in GetRequest object.
        Skips "None" values
        """
        for key, value in new_dict.items():
            if value is not None:
                self.__dict__.update({key: value})

    def enforce_types(self):
        type_enforced_dict = {}
        for key, value in self.__dict__.items():
            if key in [
                "no_directories",
                "sync",
                "sync_delete",
                "dry_run",
                "overwrite",
                "skip_existing",
                "index_parts",
            ]:
                new_value = bool(value) if value is not None else None
            else:
                new_value = str(value) if value else None
            type_enforced_dict[key] = new_value
        self.__dict__.update(type_enforced_dict)

    def from_file(self, filepath: pathlib.Path):
        json_file = json.load(open(filepath))
        json_with_mapped_options = {}
        for key, val in json_file.items():
            if key in MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS:
                new_key = MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS[key]
                json_with_mapped_options[new_key] = val
            else:
                json_with_mapped_options[key] = val
        self.__dict__.update(json_with_mapped_options)
        self.enforce_types()
        full_regex = self.regex
        if self.filter:
            filter_regex = filter_to_regex(self.filter)
            full_regex = overload_regex_with_additionnal_filter(
                filter_regex, full_regex
            )
        if self.file_list:
            file_list_regex = file_list_to_regex(self.file_list)
            full_regex = overload_regex_with_additionnal_filter(
                file_list_regex, full_regex
            )
        self.regex = full_regex


def filter_to_regex(filter: str) -> str:
    return fnmatch.translate(filter)


def file_list_to_regex(file_list_path: pathlib.Path) -> str:
    pattern = ""
    with open(file_list_path) as file_list:
        pattern = "|".join(map(re.escape, file_list.read().splitlines()))
    return pattern


def overload_regex_with_additionnal_filter(
    regex: str, filter: Optional[str]
) -> str:
    return "(" + regex + "|" + filter + ")" if filter else regex



import multiprocessing

multiprocessing.freeze_support()

import click

from copernicusmarine.command_line_interface.group_describe import cli_describe
from copernicusmarine.command_line_interface.group_get import cli_get
from copernicusmarine.command_line_interface.group_login import cli_login
from copernicusmarine.command_line_interface.group_subset import cli_subset

multiprocessing.freeze_support()  # TODO: check if we can remove some of the calls


@click.command(
    cls=click.CommandCollection,
    sources=[
        cli_describe,
        cli_login,
        cli_subset,
        cli_get,
    ],
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(None, "-V", "--version", package_name="copernicusmarine")
def base_command_line_interface():
    pass


def command_line_interface():
    base_command_line_interface(windows_expand_args=False)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    command_line_interface()



import pathlib
from typing import Optional, Union

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.exceptions import (
    MutuallyExclusiveArguments,
)
from copernicusmarine.core_functions.get import get_function
from copernicusmarine.core_functions.models import ResponseGet
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def get(
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    no_directories: bool = False,
    output_directory: Optional[Union[pathlib.Path, str]] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    overwrite: bool = False,
    request_file: Optional[Union[pathlib.Path, str]] = None,
    filter: Optional[str] = None,
    regex: Optional[str] = None,
    file_list: Optional[Union[pathlib.Path, str]] = None,
    create_file_list: Optional[str] = None,
    index_parts: bool = False,
    sync: bool = False,
    sync_delete: bool = False,
    skip_existing: bool = False,
    dry_run: bool = False,
    max_concurrent_requests: int = 15,
    disable_progress_bar: bool = False,
    staging: bool = False,
) -> ResponseGet:
    """
    Download originally produced data files.

    The datasetID is required (either as an argument or in a request file) and can be found via the ``describe`` command.

    Parameters
    ----------
    dataset_id : str, optional
        The datasetID, required either as an argument or in the request_file option.
    dataset_version : str, optional
        Force the selection of a specific dataset version.
    dataset_part : str, optional
        Force the selection of a specific dataset part.
    username : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    password : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    no_directories : bool, optional
        If True, downloaded files will not be organized into directories.
        Mutually exclusive with ``sync_delete``.
    output_directory : Union[pathlib.Path, str], optional
        The destination folder for the downloaded files. Default is the current directory.
    credentials_file : Union[pathlib.Path, str], optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files.
    overwrite : bool, optional
        If specified and if the file already exists on destination, then it will be overwritten. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
        Mutually exclusive with ``skip_existing``, ``sync`` and ``sync_delete``.
    request_file : Union[pathlib.Path, str], optional
        Option to pass a file containing the arguments. For more information please refer to the documentation or use option ``--create-template`` from the command line interface for an example template.
    filter : str, optional
        A pattern that must match the absolute paths of the files to download.
    regex : str, optional
        The regular expression that must match the absolute paths of the files to download.
    file_list : Union[pathlib.Path, str], optional
        Path to a '.txt' file containing a list of file paths, line by line, that will be downloaded directly. These files must be from the same dataset as the one specified dataset with the datasetID option. If no files can be found, the Toolbox will list all files on the remote server and attempt to find a match.
    create_file_list : str, optional
        Option to only create a file containing the names of the targeted files instead of downloading them. It writes the file to the specified output directory (default to current directory). The file name specified should end with '.txt' or '.csv'. If specified, no other action will be performed.
    sync : bool, optional
        Option to synchronize the local directory with the remote directory. See the documentation for more details. Requires to set ``dataset_version``.
        Mutually exclusive with ``skip_existing`` and ``overwrite``.
    sync_delete : bool, optional
        Option to delete local files that are not present on the remote server while applying sync. Requires to set ``dataset_version``.
        Mutually exclusive with ``skip_existing``, ``overwrite`` and ``no_directories``.
    skip_existing : bool, optional
        If the files already exists where it would be downloaded, then the download is skipped for this file. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
        Mutually exclusive with ``overwrite``, ``sync`` and ``sync_delete``.
    index_parts : bool, optional
        Option to get the index files of an INSITU dataset.
    dry_run : bool, optional
        If True, runs query without downloading data.
    max_concurrent_requests : int, optional
        Maximum number of concurrent requests. Default 15. The command uses a thread pool executor to manage concurrent requests. If set to 0, no parallel executions are used.
    disable_progress_bar : bool, optional
        Flag to hide progress bar.

    Returns
    -------
    ResponseGet
        A list of files that were downloaded and some metadata.

    """  # noqa
    # Mutually exclusive options:
    if overwrite:
        if skip_existing:
            raise MutuallyExclusiveArguments("overwrite", "skip_existing")
        elif sync:
            raise MutuallyExclusiveArguments("overwrite", "sync")
        elif sync_delete:
            raise MutuallyExclusiveArguments("overwrite", "sync_delete")
    if skip_existing:
        if sync:
            raise MutuallyExclusiveArguments("skip_existing", "sync")
        elif sync_delete:
            raise MutuallyExclusiveArguments("skip_existing", "sync_delete")
    if no_directories:
        if sync_delete:
            raise MutuallyExclusiveArguments("sync_delete", "no_directories")

    output_directory = (
        pathlib.Path(output_directory) if output_directory else None
    )
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )
    file_list = pathlib.Path(file_list) if file_list else None
    request_file = pathlib.Path(request_file) if request_file else None
    return get_function(
        dataset_id=dataset_id,
        force_dataset_version=dataset_version,
        force_dataset_part=dataset_part,
        username=username,
        password=password,
        no_directories=no_directories,
        output_directory=output_directory,
        credentials_file=credentials_file,
        overwrite=overwrite,
        request_file=request_file,
        filter_option=filter,
        regex=regex,
        file_list_path=file_list,
        create_file_list=create_file_list,
        index_parts=index_parts,
        sync=sync,
        sync_delete=sync_delete,
        skip_existing=skip_existing,
        dry_run=dry_run,
        max_concurrent_requests=max_concurrent_requests,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )



from copernicusmarine.core_functions.models import CommandType

SHARED: dict[str, str] = {
    "OVERWRITE_HELP": "If specified and if the file already exists on destination, then it will be overwritten. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').",  # noqa: E501
    "USERNAME_HELP": (
        "If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, then search for a credentials file, else ask for user input."  # noqa
    ),  # a little hardcoding in Python API
    "PASSWORD_HELP": (
        "If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, then search for a credentials file, else ask for user input."  # noqa
    ),  # a little hardcoding in Python API
    "LOG_LEVEL_HELP": (
        "Set the details printed to console by the command "
        "(based on standard logging library)."
    ),
    "CREATE_TEMPLATE_HELP": (
        "Option to create a file <argument>_template.json in your current directory "
        "containing the arguments. If specified, no other action will be performed."
    ),
    "CREDENTIALS_FILE_HELP": (
        "Path to a credentials file if not in its default directory"
        " (``$HOME/.copernicusmarine``). Accepts "
        ".copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini "
        "files."
    ),
    "DATASET_VERSION_HELP": "Force the selection of a specific dataset version.",
    "DATASET_PART_HELP": "Force the selection of a specific dataset part.",
    "DATASET_ID_HELP": (
        "The datasetID, required either as an argument or in the request_file option."
    ),
    "DISABLE_PROGRESS_BAR_HELP": "Flag to hide progress bar.",
    "DRY_RUN_HELP": "If True, runs query without downloading data.",
    "RESPONSE_FIELDS_HELP": (
        "List of fields to include in the query metadata. "
        "The fields are separated by a comma. "
        "To return all fields, use 'all'."
    ),
    "OUTPUT_DIRECTORY_HELP": (
        "The destination folder for the downloaded files. Default is the current "
        "directory."
    ),
    "REQUEST_FILE_HELP": (
        "Option to pass a file containing the arguments. For more information "
        "please refer to the documentation or use option ``--create-template`` "
        "from the command line interface for an example template."
    ),
    "SKIP_EXISTING_HELP": (
        "If the files already exists where it would be downloaded, then "
        "the download is skipped for this file. By default, the toolbox "
        "creates a new file with a new index (eg 'filename_(1).nc')."
    ),
}

LOGIN: dict[str, str] = {
    "LOGIN_DESCRIPTION_HELP": (
        "Create a configuration file with your Copernicus Marine credentials"
        " under the ``$HOME/.copernicusmarine`` directory."
    ),
    "LOGIN_RESPONSE_HELP": (
        "Exit code\n 0 if the login was successfully completed, 1 otherwise."
    ),
    "USERNAME_HELP": (
        "If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, else ask for user input."  # noqa
    ),
    "PASSWORD_HELP": (
        "If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, else ask for user input."  # noqa
    ),
    "CONFIGURATION_FILE_DIRECTORY_HELP": (
        "Path to the directory where the configuration file will be stored."
    ),
    "FORCE_OVERWRITE_HELP": (
        "Flag to skip confirmation before overwriting configuration file."
    ),
    "CHECK_CREDENTIALS_VALID_HELP": (
        "Flag to check if the credentials are valid. "
        "No other action will be performed. "
        "The validity will be check in this order: "
        "1. Check if the credentials are valid"
        " with the provided username and password. "
        "2. Check if the credentials are valid in the environment variables. "
        "3. Check if the credentials are valid in the configuration file. "
        "When any is found (valid or not valid), will return immediately."
    ),
    "CREDENTIALS_FILE_HELP": (
        "Path to a credentials file if not in its default directory"
        " (``$HOME/.copernicusmarine``). Accepts "
        ".copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini "
        "files. Will only be taken into account when checking the credentials validity."
    ),
}

DESCRIBE: dict[str, str] = {
    "DESCRIBE_DESCRIPTION_HELP": (
        "Retrieve and parse the metadata information "
        "from the Copernicus Marine catalogue."
    ),
    "DESCRIBE_RESPONSE_HELP": (
        "JSON\n A dictionary containing the retrieved metadata information."
    ),
    "SHOW_ALL_VERSIONS_HELP": (
        "Include dataset versions in output. By default, shows only the default "
        "version."
    ),
    "RETURN_FIELDS_HELP": (
        "Option to specify the fields to return in the output. "
        "The fields are separated by a comma. You can use 'all' to return all fields."
    ),
    "EXCLUDE_FIELDS_HELP": (
        "Option to specify the fields to exclude from the output. "
        "The fields are separated by a comma."
    ),
    "CONTAINS_HELP": (
        "Filter catalogue output. Returns products with attributes matching a string "
        "token."
    ),
    "PRODUCT_ID_HELP": (
        "Force the productID to be used for the describe command. Will not parse the "
        "whole catalogue, but only the product with the given productID."
    ),
    "DATASET_ID_HELP": (
        "Force the datasetID to be used for the describe command. Will not "
        "parse the whole catalogue, but only the dataset with the given datasetID."
    ),
    "MAX_CONCURRENT_REQUESTS_HELP": (
        "Maximum number of concurrent requests (>=1). Default 15. The command uses "
        "a thread pool executor to manage concurrent requests."
    ),
    "RAISE_ON_ERROR_HELP": (
        "If set to True, the function will raise at the first error encountered "
        "during the parsing and fetching of the catalogue. Default, False."
    ),
}

SUBSET: dict[str, str] = {
    "SUBSET_DESCRIPTION_HELP": (
        "Extract a subset of data from a specified dataset using given parameters."
        "\n\nThe datasetID is required and can be found via the ``describe`` "
        "command. "  # has some hardcoding in CLI and python API
    ),
    "SUBSET_RESPONSE_HELP": (
        "JSON \n A description of the downloaded data and its destination."
    ),
    "SERVICE_HELP": (
        f"Force download through one of the available services using the service name "
        f"among {CommandType.SUBSET.service_names()} or "
        f"its short name among {CommandType.SUBSET.short_names_services()}."
    ),
    "VARIABLES_HELP": "Specify dataset variable. Can be used multiple times.",
    "MINIMUM_LONGITUDE_HELP": (
        "Minimum longitude for the subset. The value will be transposed "
        "to the interval [-180; 360[."
    ),
    "MINIMUM_X_HELP": (
        "Minimum x-axis value for the subset. "
        "The units are considered in length (m, 100km...)."
    ),
    "ALIAS_MIN_X_HELP": (
        "Alias for ``--minimum-longitude`` and ``--minimum-x``."
    ),
    "MAXIMUM_LONGITUDE_HELP": (
        "Maximum longitude for the subset. The value will be transposed"
        " to the interval [-180; 360[."
    ),
    "ALIAS_MAX_X_HELP": (
        "Alias for ``--maximum-longitude`` and ``--maximum-x``."
    ),
    "MAXIMUM_X_HELP": (
        "Maximum x-axis value for the subset. "
        "The units are considered in length (m, 100km...)."
    ),
    "MINIMUM_LATITUDE_HELP": (
        "Minimum latitude for the subset. Requires a float from -90 "
        "degrees to +90."
    ),
    "ALIAS_MIN_Y_HELP": (
        "Alias for ``--minimum-latitude`` and ``--minimum-y``."
    ),
    "MINIMUM_Y_HELP": (
        "Minimum y-axis value for the subset. "
        "The units are considered in length (m, 100km...)."
    ),
    "MAXIMUM_LATITUDE_HELP": (
        "Maximum latitude for the subset. Requires a float from -90 degrees "
        "to +90."
    ),
    "ALIAS_MAX_Y_HELP": (
        "Alias for ``--maximum-latitude`` and ``--maximum-y``."
    ),
    "MAXIMUM_Y_HELP": (
        "Maximum y-axis value for the subset. "
        "The units are considered in length (m, 100km...)."
    ),
    "MINIMUM_DEPTH_HELP": ("Minimum depth for the subset."),
    "MAXIMUM_DEPTH_HELP": ("Maximum depth for the subset."),
    "VERTICAL_AXIS_HELP": (
        "Consolidate the vertical dimension (the z-axis) as requested: depth with "
        "descending positive values, elevation with ascending positive values. "
        "Default is depth."
    ),
    "START_DATETIME_HELP": (
        "The start datetime of the temporal subset. Supports common "
        "format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html)."  # noqa
    ),  # hardocded in cli: Caution: encapsulate date with “ “ to ensure valid
    # expression for format “%Y-%m-%d %H:%M:%S”.
    "END_DATETIME_HELP": (
        "The end datetime of the temporal subset. Supports common "
        "format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html)."  # noqa
    ),  # hardocded in cli: Caution: encapsulate date with “ “
    # to ensure valid expression for format “%Y-%m-%d %H:%M:%S”.
    "PLATFORM_IDS_HELP": (
        "Specify platform ID. Can be used multiple times. "
        "Only available for platform chunked datasets."
    ),
    "COORDINATES_SELECTION_METHOD_HELP": (
        "If ``inside``, the "
        "selection retrieved will be inside the requested range. If ``strict-"
        "inside``, the selection retrieved will be inside the requested range, "
        "and an error will be raised if the values don't exist. "
        "If ``nearest``, the extremes closest to the requested values will "
        "be returned. If ``outside``,"
        " the extremes will be taken to contain all the requested interval."
        " The methods ``inside``, ``nearest`` and ``outside`` will display"
        " a warning if the request is out of bounds."
    ),
    "OUTPUT_FILENAME_HELP": (
        "Save the downloaded data with the given file name (under the output "
        "directory)."
    ),
    "FILE_FORMAT_HELP": "Format of the downloaded dataset. If not set or set to ``None``, defaults to NetCDF '.nc' for gridded datasets and to CSV '.csv' for sparse datasets.",  # noqa: E501
    "MOTU_API_REQUEST_HELP": (
        "Option to pass a complete MOTU API request as a string. Caution, user has to "
        """replace double quotes " with single quotes ' in the request."""
    ),
    "NETCDF_COMPRESSION_LEVEL_HELP": (
        "Specify a compression level to apply on the NetCDF output file. A value of 0 "
        "means no compression, and 9 is the highest level of compression available."
    ),  # some hardcoding in CLI to add the flag value
    "NETCDF3_COMPATIBLE_HELP": (
        "Enable downloading the dataset in a netCDF3 compatible format."
    ),
    "CHUNK_SIZE_LIMIT_HELP": (
        "Limit the size of the chunks in the dask array. Default is set to -1 which "
        "behaves similarly to 'chunks=auto' from ``xarray``. Positive integer"
        " values and '-1' are accepted. This is an experimental feature."
    ),
    "RAISE_IF_UPDATING_HELP": (
        "If set, raises a :class:`copernicusmarine.DatasetUpdating` "
        "error if the dataset is being updated "
        "and the subset interval requested overpasses "
        "the updating start date of the dataset."
        " Otherwise, a simple warning is displayed."
    ),
    "SPLIT_ON_HELP": (
        "When subsetting to netCDF, allows to split the output "
        "file by variable or by time. This option also "
        "enables parallel downloading for netCDF. "
        "See :ref:`split-on option <subset-split-on>` in the documentation."
    ),
}

GET: dict[str, str] = {
    "GET_DESCRIPTION_HELP": (
        "Download originally produced data files.\n\n"
        "The datasetID is required (either as an "
        "argument or in a request file) and can be found via the ``describe``"
        " command."
    ),  # has some hardcoding in CLI
    "MAX_CONCURRENT_REQUESTS_HELP": (
        "Maximum number of concurrent requests. Default 15. The command uses a thread "
        "pool executor to manage concurrent requests. If set to 0, no parallel"
        " executions are used."
    ),
    "GET_RESPONSE_HELP": (
        "JSON \n A list of files that were downloaded and some metadata."
    ),
    "FILTER_HELP": (
        "A pattern that must match the absolute paths of the files to download."
    ),
    "REGEX_HELP": (
        "The regular expression that must match the absolute paths of the files to "
        "download."
    ),
    "FILE_LIST_HELP": (
        "Path to a '.txt' file containing a "
        "list of file paths, line by line, that will "
        "be downloaded directly. These files must be from the same dataset as the one s"
        "pecified dataset with the datasetID option. If no files can be found, the "
        "Toolbox will list all files on the remote server and attempt to find a match."
    ),
    "CREATE_FILE_LIST_HELP": (
        "Option to only create a file containing the names of the targeted files "
        "instead of downloading them. It writes the file to the specified output"
        " directory (default to current directory). The file "
        "name specified should end with '.txt' or '.csv'. If specified, no other "
        "action will be performed."
    ),
    "SYNC_HELP": (
        "Option to synchronize the local directory with the remote directory. See the "
        "documentation for more details. Requires to set ``--dataset-version``."
    ),
    "SYNC_DELETE_HELP": (
        "Option to delete local files that are not present on the remote server while "
        "applying sync. Requires to set ``--dataset-version``."
    ),
    "INDEX_PARTS_HELP": (
        "Option to get the index files of an INSITU dataset."
    ),
    "NO_DIRECTORIES_HELP": (
        "If True, downloaded files will not be organized into directories."
    ),
}

SUBSET_SPLIT_ON: dict[str, str] = {
    "SPLIT_ON_DESCRIPTION_HELP": (
        "Extract a subset of data from a specified dataset using given parameters "
        "and split the output files based on variable names, time intervals or both. "
        "By default, sequentially download the data.\n\n"
        "The datasetID is required and can be found via the ``describe`` command. "
        "Accept all the arguments from the subset, and should be entered by order: "
        "``copernicusmarine subset [SUBSET-OPTIONS] split-on [SPLIT-ON-OPTIONS].``"
    ),
    "SPLIT_ON_SHORT_HELP": "Extract a subset of data and split the output files.",
    "ON_VARIABLES_HELP": (
        "If True, split the output files based on variable names, by default False."
    ),
    "ON_TIME_HELP": (
        "If provided, split the output files based on specified time intervals, "
        "by default None."
    ),
    "CONCURRENT_PROCESSES_HELP": (
        "Number of concurrent processes to use for downloading data, "
        "by default None. Should be greater or equal to 1."
    ),
}

SUBSET.update(SHARED)
GET.update(SHARED)
LOGIN.update({k: v for k, v in SHARED.items() if k not in LOGIN})
DESCRIBE.update({k: v for k, v in SHARED.items() if k not in DESCRIBE})



import xarray

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.exceptions import (
    FormatNotSupported,
    ServiceNotSupported,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.services_utils import RetrievalService
from copernicusmarine.core_functions.subset import (
    retrieve_metadata_and_check_request,
)
from copernicusmarine.download_functions.download_zarr import (
    get_dataset_and_parameters,
)


def open_dataset_function(
    subset_request: SubsetRequest,
) -> xarray.Dataset:
    retrieval_service: RetrievalService = retrieve_metadata_and_check_request(
        subset_request
    )
    if (
        retrieval_service.service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ):
        raise FormatNotSupported(
            CopernicusMarineServiceFormat.SQLITE.value,
            "open_dataset",
            "read_dataframe",
        )
    if retrieval_service.service_name not in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        raise ServiceNotSupported(retrieval_service.service_name)

    dataset, _, _ = get_dataset_and_parameters(
        subset_request=subset_request,
        dataset_url=retrieval_service.uri,
        axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
        service=retrieval_service.service,
        dataset_chunking=retrieval_service.dataset_chunking,
        is_original_grid=retrieval_service.is_original_grid,
        dataset_valid_start_date=retrieval_service.dataset_valid_start_date,
    )

    return dataset



import logging
import pathlib
from typing import Optional, Union

import click

from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.command_line_interface.utils import (
    MutuallyExclusiveOption,
    assert_cli_args_are_not_set_except_create_template,
    credentials_file_option,
    force_dataset_part_option,
    force_dataset_version_option,
    force_download_option,
    tqdm_disable_option,
)
from copernicusmarine.core_functions import documentation_utils
from copernicusmarine.core_functions.click_custom_class import (
    CustomClickOptionsCommand,
)
from copernicusmarine.core_functions.fields_query_builder import (
    build_query,
    get_queryable_requested_fields,
)
from copernicusmarine.core_functions.get import (
    create_get_template,
    get_function,
)
from copernicusmarine.core_functions.models import ResponseGet

logger = logging.getLogger("copernicusmarine")
blank_logger = logging.getLogger("copernicusmarine_blank_logger")

DEFAULT_FIELDS_TO_INCLUDE = {
    "status",
    "message",
    "total_size",
    "number_of_files_to_download",
}


@click.group()
def cli_get() -> None:
    pass


@cli_get.command(
    "get",
    cls=CustomClickOptionsCommand,
    short_help="Download originally produced data files.",
    help=documentation_utils.GET["GET_DESCRIPTION_HELP"]
    + " See :ref:`describe <cli-describe>`. \n\nReturns\n "
    + documentation_utils.GET["GET_RESPONSE_HELP"],
    epilog="""
    Example to download all the files from a given dataset:

    .. code-block:: bash

        copernicusmarine get -i cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m \n
    """,  # noqa
)
@click.option(
    "--dataset-id",
    "-i",
    type=str,
    default=None,
    help=documentation_utils.GET["DATASET_ID_HELP"],
)
@force_dataset_version_option
@force_dataset_part_option
@click.option(
    "--username",
    type=str,
    default=None,
    help=documentation_utils.GET["USERNAME_HELP"],
)
@click.option(
    "--password",
    type=str,
    default=None,
    help=documentation_utils.GET["PASSWORD_HELP"],
)
@click.option(
    "--no-directories",
    "-nd",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help=documentation_utils.GET["NO_DIRECTORIES_HELP"],
    default=False,
    mutually_exclusive=["sync-delete"],
)
@click.option(
    "--output-directory",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help=documentation_utils.GET["OUTPUT_DIRECTORY_HELP"],
)
@credentials_file_option
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    cls=MutuallyExclusiveOption,
    help=documentation_utils.GET["OVERWRITE_HELP"],
    mutually_exclusive=[
        "skip-existing",
        "sync",
        "sync-delete",
    ],
)
@click.option(
    "--create-template",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["CREATE_TEMPLATE_HELP"],
)
@click.option(
    "--request-file",
    type=click.Path(exists=True, path_type=pathlib.Path),
    help=documentation_utils.GET["REQUEST_FILE_HELP"],
)
@click.option(
    "--filter",
    "--filter-with-globbing-pattern",
    type=str,
    default=None,
    help=documentation_utils.GET["FILTER_HELP"],
)
@click.option(
    "--regex",
    "--filter-with-regular-expression",
    type=str,
    default=None,
    help=documentation_utils.GET["REGEX_HELP"],
)
@click.option(
    "--file-list",
    type=pathlib.Path,
    default=None,
    help=documentation_utils.GET["FILE_LIST_HELP"],
)
@click.option(
    "--create-file-list",
    type=str,
    default=None,
    help=documentation_utils.GET["CREATE_FILE_LIST_HELP"],
)
@click.option(
    "--sync",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["SYNC_HELP"],
    mutually_exclusive=["skip-existing", "overwrite"],
)
@click.option(
    "--sync-delete",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["SYNC_DELETE_HELP"],
    mutually_exclusive=[
        "no-directories",
        "overwrite",
        "skip-existing",
    ],
)
@click.option(
    "--skip-existing",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["SKIP_EXISTING_HELP"],
    mutually_exclusive=[
        "overwrite",
        "sync",
        "sync-delete",
    ],
)
@click.option(
    "--index-parts",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["INDEX_PARTS_HELP"],
)
@click.option(
    "--dry-run",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.GET["DRY_RUN_HELP"],
)
@click.option(
    "--response-fields",
    "-r",
    type=str,
    default=None,
    help=documentation_utils.GET["RESPONSE_FIELDS_HELP"],
)
@click.option(
    "--max-concurrent-requests",
    type=int,
    default=15,
    help=documentation_utils.GET["MAX_CONCURRENT_REQUESTS_HELP"],
)
@tqdm_disable_option
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=documentation_utils.GET["LOG_LEVEL_HELP"],
)
@click.option(
    "--staging",
    type=bool,
    default=False,
    is_flag=True,
    hidden=True,
)
@force_download_option
@log_exception_and_exit
def get(
    dataset_id: Optional[str],
    dataset_version: Optional[str],
    dataset_part: Optional[str],
    username: Optional[str],
    password: Optional[str],
    no_directories: bool,
    output_directory: Optional[pathlib.Path],
    credentials_file: Optional[pathlib.Path],
    overwrite: bool,
    create_template: bool,
    request_file: Optional[pathlib.Path],
    filter: Optional[str],
    regex: Optional[str],
    file_list: Optional[pathlib.Path],
    create_file_list: Optional[str],
    sync: bool,
    sync_delete: bool,
    skip_existing: bool,
    index_parts: bool,
    dry_run: bool,
    response_fields: Optional[str],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    log_level: str,
    staging: bool,
    force_download: bool,
):
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUG mode activated")

    if create_template:
        assert_cli_args_are_not_set_except_create_template(
            click.get_current_context()
        )
        create_get_template()
        return

    response = get_function(
        dataset_id=dataset_id,
        force_dataset_version=dataset_version,
        force_dataset_part=dataset_part,
        username=username,
        password=password,
        no_directories=no_directories,
        output_directory=output_directory,
        credentials_file=credentials_file,
        overwrite=overwrite,
        request_file=request_file,
        filter_option=filter,
        regex=regex,
        file_list_path=file_list,
        create_file_list=create_file_list,
        sync=sync,
        sync_delete=sync_delete,
        skip_existing=skip_existing,
        index_parts=index_parts,
        dry_run=dry_run,
        max_concurrent_requests=max_concurrent_requests,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )

    if response_fields:
        fields_to_include = set(response_fields.replace(" ", "").split(","))
    elif dry_run:
        fields_to_include = {"all"}
    else:
        fields_to_include = DEFAULT_FIELDS_TO_INCLUDE
    included_fields: Optional[Union[set[str], dict]]
    if "all" in fields_to_include:
        included_fields = None
    elif "none" in fields_to_include:
        included_fields = set()
    else:
        queryable_fields = get_queryable_requested_fields(
            fields_to_include, ResponseGet, "--response-fields"
        )
        included_fields = build_query(set(queryable_fields), ResponseGet)

    blank_logger.info(
        response.model_dump_json(
            indent=2,
            include=included_fields,
            exclude_none=True,
            exclude_unset=True,
        )
    )



import functools
import logging
from collections.abc import Iterator, Mapping
from typing import Any, Callable

logger = logging.getLogger("copernicusmarine")


class DeprecatedOption:
    def __init__(
        self, old_name, new_name, replace=True, do_not_pass=False
    ) -> None:
        self.old_name = old_name
        self.new_name = new_name
        self.replace = replace
        self.do_not_pass = do_not_pass


class DeprecatedOptionMapping(Mapping):
    def __init__(self, deprecated_options: list[DeprecatedOption]) -> None:
        self.deprecated_options_by_old_names: dict[str, DeprecatedOption] = {}
        for value in deprecated_options:
            if value not in self.deprecated_options_by_old_names:
                self.deprecated_options_by_old_names[value.old_name] = value

    def __getitem__(self, __key: str) -> DeprecatedOption:
        return self.deprecated_options_by_old_names[__key]

    def __iter__(self) -> Iterator:
        return self.deprecated_options_by_old_names.__iter__()

    def __len__(self) -> int:
        return self.deprecated_options_by_old_names.__len__()


DEPRECATED_OPTIONS: DeprecatedOptionMapping = DeprecatedOptionMapping(
    [
        DeprecatedOption(
            old_name="motu_api_request",
            new_name="motu_api_request",
            replace=False,
        ),
        DeprecatedOption(
            old_name="force_download",
            new_name="force_download",
            replace=False,
            do_not_pass=True,
        ),
    ]
)


def get_deprecated_message(old_value, preferred_value):
    return (
        f"'{old_value}' has been deprecated, use '{preferred_value}' instead."
    )


def log_deprecated_message(old_value, preferred_value):
    if preferred_value is None:
        logger.warning(f"'{old_value}' has been deprecated.")
    else:
        logger.warning(get_deprecated_message(old_value, preferred_value))


def raise_both_old_and_new_value_error(old_value, new_value):
    raise TypeError(
        f"Received both {old_value} and {new_value} as arguments! "
        f"{get_deprecated_message(old_value, new_value)}"
    )


def deprecated_python_option(aliases: DeprecatedOptionMapping) -> Callable:
    def deco(f: Callable):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(
    func_name: str, kwargs: dict[str, Any], aliases: DeprecatedOptionMapping
):
    for old, alias_info in aliases.deprecated_options_by_old_names.items():
        new = alias_info.new_name
        if old in kwargs:
            if new in kwargs and old != new:
                raise_both_old_and_new_value_error(old, new)
            if old == new:
                log_deprecated_message(old, None)
            else:
                log_deprecated_message(old, new)
            if alias_info.replace:
                kwargs[new] = kwargs.pop(old)
            if alias_info.do_not_pass:
                del kwargs[old]



import inspect
import logging
import sys

import click
from click import Context

from copernicusmarine.core_functions.deprecated_options import (
    log_deprecated_message,
)

logger = logging.getLogger("copernicusmarine")


class CustomDeprecatedClickOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.custom_deprecated = kwargs.pop("custom_deprecated", ())
        self.preferred = kwargs.pop("preferred", None)
        super().__init__(*args, **kwargs)


class CustomClickOptionsCommand(click.Command):
    def make_parser(self, ctx: Context):
        parser = super().make_parser(ctx)

        # get the parser options
        options = set(parser._short_opt.values())
        options |= set(parser._long_opt.values())

        for option in options:
            if not isinstance(option.obj, CustomDeprecatedClickOption):
                continue

            def make_process(an_option):
                orig_process = an_option.process
                custom_deprecated = getattr(
                    an_option.obj, "custom_deprecated", None
                )
                preferred = getattr(an_option.obj, "preferred", None)
                msg = "Expected `deprecated` value for `{}`"
                assert custom_deprecated is not None, msg.format(
                    an_option.obj.name
                )

                def process(value, state):
                    frame = inspect.currentframe()
                    try:
                        if frame and frame.f_back:
                            opt = frame.f_back.f_locals.get("opt")
                    finally:
                        del frame

                    if opt in custom_deprecated:
                        log_deprecated_message(opt, preferred)
                    return orig_process(value, state)

                return process

            option.process = make_process(option)

        return parser

    def format_epilog(self, ctx, formatter):
        if self.epilog:
            formatter.write_paragraph()
            for line in self.epilog.split("\n"):
                if ".. code-block::" in line:
                    continue
                formatter.write(line + "\n")


class CustomClickOptionsGroup(click.Group):
    def make_parser(self, ctx):
        parser = super().make_parser(ctx)

        options = set(parser._short_opt.values())
        options |= set(parser._long_opt.values())

        # so that we can skip validation of the subset options
        # in case help is requested for split-on
        target = "split-on"
        argv = sys.argv[1:]
        ctx.meta["help_for"] = None
        if target in argv:
            ctx.meta["command"] = target
            idx = argv.index(target)
            after = argv[idx + 1 :]
            help_flags = {"-h", "--help"}
            if any(token in help_flags for token in after):
                ctx.meta["help_for"] = target

        for option in options:
            if not isinstance(option.obj, CustomDeprecatedClickOption):
                continue

            def make_process(an_option):
                orig_process = an_option.process
                custom_deprecated = getattr(
                    an_option.obj, "custom_deprecated", None
                )
                preferred = getattr(an_option.obj, "preferred", None)
                msg = "Expected `deprecated` value for `{}`"
                assert custom_deprecated is not None, msg.format(
                    an_option.obj.name
                )

                def process(value, state):
                    frame = inspect.currentframe()
                    try:
                        if frame and frame.f_back:
                            opt = frame.f_back.f_locals.get("opt")
                    finally:
                        del frame

                    if opt in custom_deprecated:
                        log_deprecated_message(opt, preferred)
                    return orig_process(value, state)

                return process

            option.process = make_process(option)

        return parser

    def format_epilog(self, ctx, formatter):
        if self.epilog:
            formatter.write_paragraph()
            for line in self.epilog.split("\n"):
                if ".. code-block::" in line:
                    continue
                formatter.write(line + "\n")
